package Config;
use NBK::Planner::Planner;
use NBK::Planner::Plan;

=head1 NAME

Config  - NodeBrain Kit Configuration Module

=head1 SYNOPSIS

  use lib "lib";  # We assume the working directory is the application directory
  use NBK::Config;
  &Config::install($plan,$component);     // install a kit component

=head1 DESCRIPTION

  &Config::install($plan,$component);

  $plan   - Monitor name to be displayed in messages.  This should
               correspond to the configuration file <plan>.cfg being
               processed.                  
  
  $component - Name of component to install, qualified relative to the
               caboodle. (e.g. "adapter/threshold.config")
  
  Exit Codes:
    0 - successful
    1 - unsuccessful

  Change History:

  2001/04/15 eat - version 0.7.5
             1) Modified load to prefer ";" as field separator but still
                support ":".
  2001/04/16 eat - version 0.7.5
             1) Now support ")entry" sections              
  2002/02/13 eat - version 0.8.0
             1) Now supports ")let" sections for symbolic substitution        
                    )let
                    variable=value

                    )case ...
                    .......${variable}.....
  2007/06/13 eat 0.6.7  Switched to NodeBrain Kit version numbers
  2007/06/13 eat 0.6.7  Using relative paths from caboodle
  2007/11/11 eat 0.6.8  Included support for )code
                    )code
                    ... nb code to eof ...
  2013-02-10 eat 0.8.13 Changed location of kit directories
            
=cut

# Get list of kits associated with a caboodle

sub getKitList{
  my ($kitDir,$kitHash)=@_;
  my $kitfile="$kitDir/kit/kits";
  return if(!(-f "$kitfile"));
  my @kit;
  my %kit;
  open(KITS,"$kitfile")||die "CAB000E Unable to open kits file $kitfile";
  while(<KITS>){
    my $kitref=$_;
    chomp($kitref);
    next if($kitref=~/^#|^\s*$/);
    my $kitDir="kit/use/$kitref";
    push(@kit,$kitDir);
    $$kitHash{$kitDir}=1;
    }
  close(KITS);
  return(reverse(@kit));
  }

# Get component from a kit directory
#   This should be moved to a common module after it is worked out
sub getKitComponent{
  my ($component,$kitdir)=@_;

  return(0) if($componentCheck{"$kitdir/$component"});               # avoid infinite recursion
  $componentCheck{"$kitdir/$component"}=1;
  if(-f "$kitdir/$component"){
    my $dir;
    $dir=$1 if($component=~/^(.*)\/([^\/]+)$/);
    system("mkdir -p $dir") if($dir && !(-d "$dir"));   # make directory if necessary
    system("cp $kitdir/$component $component");  # copy the component from the kit
    return(1);
    }
  getComponent($component,$kitdir);
  }

# Get component from kit directories

sub getComponent{
  my ($component,$kitdir)=@_;

  return(0) if(!(-f "$kitdir/kit/kits"));
  if(!open(KITS,"$kitdir/kit/kits")){
    print(STDERR "Unable to open kits file in kit $kitdir/kit\n");
    exit(1);
    }
  my @kitdir;
  while(<KITS>){
    next if(/^#/);
    chomp();
    push(@kitdir,"kit/use/$_");
    }
  close(KITS);
  while(@kitdir){
    my $kitdir=pop(@kitdir);
    return(1) if(getKitComponent($component,$kitdir));
    }
  return(0);
  }

# Install executable component

sub install{
  my ($plan,$component)=@_;
  if(! -f "$component"){
    if(!getComponent($component,".")){
      print(STDERR "CAB000E $plan: Component \"$component\" not available.\n");
      return(1);
      }
    if(-f "$component"){
      system("chmod 755 $component >/dev/null 2>\&1");
      print(STDERR "CAB000I $plan: Component \"$component\" installed.\n");
      return(0);
      }
    else{
      print(STDERR "CAB000E $plan: Component \"$component\" install failed.\n");   
      return(1);
      }
    }
  return(0);
  }
        
sub esc2oct{
  ($_)=@_;
#  print($_."\n");
  $_=join('\136',split(/\\\\/,$_)); # \/
  $_=join('\059',split(/\\\;/,$_)); # \;
  $_=join('\072',split(/\\:/,$_));  # \:
  $_=join('\047',split(/\\'/,$_));  # \'
  $_=join('\042',split(/\\"/,$_));  # \"
  return($_);
  }
    
sub getLine{
  $line=<CONFIG>;
  chomp($line);
  $line=esc2oct($line);
  while(!eof(CONFIG) && ($line=~/^#/ || $line=~/^\s*$/)){
    $line=<CONFIG>;
    chomp($line);
    $line=esc2oct($line);
    }
  return('') if($line=~/^#/ || $line=~/^\s*$/);
  $prefix='';
  while($line=~/^(.*?)\$\{(.*?)\}(.*)$/){
    if(!defined($let{$2})){
      print("CAB000E $plan: Symbolic variable \"$2\" not defined.\n");
      exit(1);
      }
    $prefix.=$1.$let{$2};
    $line=$3;
    }
  return($prefix.$line);
  }

# Get plan attributes
#
#   $msg=getPlan($plan,\%plan);
#   $msg=getPlan($plan,\%plan,$cabDir);
#
#   $cabDir - caboodle directory
#             Omit or use null string to get from current caboodle
#             Use kit caboodle directory when loading from installed kit
#
#   Fills a hash with plan attributes

sub getPlan{
  my ($plan,$hash,$cabDir)=@_;

  my $dir=$cabDir; 
  if($dir && $dir ne '.'){$dir.="/plan/$plan";}
  else{$dir="plan/$plan";}
 
  if(-f "$dir/$plan~.xml"){
    my $planner=getPlanner();
    my $msg=$planner->getCntl($plan,$hash,$cabDir);
    return($msg) if($msg);
    return('');
    }
  return("File $dir/$plan~.xml not found.");
  }

#=============================================================================
# Plan state and relationship functions
#=============================================================================

# Get related plans and the state of the relationship
#
#    plan/<plan>/<reletive>.disabled              disabled
#    plan/<relative>/<plan>.disabled              disputed
#    plan/<plan>/<relative>.child
#      plan/<relative>/<plan>.parent              enabled
#    plan/<plan>/<relative>.parent
#      plan/<relative>/<plan>.child               enabled
#    plan/<plan>                                  unilateral
#    else                                         broken
#
sub getRelated{
  my ($plan,$relationship,$relative)=@_;
  return(0) if(!opendir(DIR,"plan/$plan"));
  my @file=readdir(DIR);
  closedir(DIR);
  foreach my $file (@file){
    next if(!($file=~/^(.*)\.(parent|child)$/));
    my $name=$1;
    my $relation=$2;
    next if($relationship ne $relation);
    my $state='broken';
    my $reflection='child';
    $reflection='parent' if($relation eq 'child');
    if(-f "plan/$plan/$name.disabled"){$state='disabled';}
    elsif(-f "plan/$name/$plan.disabled" || -f "plan/$name/$name.disabled"){$state='disputed';}
    elsif(-f "plan/$name/$plan.$reflection"){$state='enabled';}
    elsif(-d "plan/$name"){$state='unilateral';} 
    $$relative{$name}=$state;
    }
  return(1);
  }

sub getParents{
  my ($plan,$relative)=@_;
  return(0) if(!opendir(DIR,"plan/$plan"));
  return(getRelated($plan,'parent',$relative));
  }

# Get children in a hash

sub getChildren{
  my ($plan,$relative)=@_;
  return(0) if(!opendir(DIR,"plan/$plan"));
  return(getRelated($plan,'child',$relative));
  }

sub getRelatedTree{
  my ($plan,$relationship,$relative)=@_;
  my %plan;
  getRelated($plan,$relationship,\%plan);
  foreach my $plan (sort keys %plan){
    if(!$$relative{$plan}){
      $$relative{$plan}=$plan{$plan};
      getRelatedTree($plan,$relationship,$relative);
      }
    }
  }

# Check to avoid creation of circular relationships

sub checkRelative{
  my ($plan,$relative,$relationship)=@_;
  my %plan;
  if($relationship eq 'parent'){
    getRelatedTree($plan,'child',\%plan);
    }
  elsif($relationship eq 'child'){
    getRelatedTree($plan,'parent',\%plan);
    }
  else{return("Relationship '$relationship' not recognized.");}
  return("Making '$relative' a $relationship of '$plan' would create a circular relationship.") if($plan{$relative});
  return('');
  }

sub getRelatives{
  my ($plan,$relation,$states)=@_;
  my %want;
  my %relative;
  my @relative;
  foreach my $state (split(/,/,$states)){
    $want{$state}=1;
    }
  if($relation ne 'children'){getParents($plan,\%relative);}
  if($relation ne 'parents'){getChildren($plan,\%relative);}
  foreach my $relative (sort keys %relative){
    push(@relative,$relative) if($want{$relative{$relative}});
    }
  return(@relative);
  }
 
# Rename a file that may be under subversion control
#
sub svnRename{
  my ($dir,$oldName,$newName)=@_;
  if(-d "$dir/.svn"){  # if under subversion control
    system("cd $dir;svn move $oldName $newName");
    return if(! -e "$dir/$newName");
    system("cd $dir;rm -fr $oldName") if(-d "$dir/$oldName");
    system("cd $dir;svn commit -m 'rename by planner'");
    }
  else{rename("$dir/$oldName","$dir/$newName");}
  }

# Rename a plan
#
# Returns:  $msg - null if successful
#
sub renamePlan{
  my ($oldName,$newName)=@_;
  #print("oldname=$oldName,newname=$newName<br>\n");
  return("Plan '$oldName' not found.") if(! -d "plan/$oldName");
  return("Plan '$newName' already exists.") if(-d "plan/$newName");
  # Rename references in parent plans
  my %parents;
  getParents($oldName,\%parents);
  foreach my $parent (keys %parents){
    #rename("plan/$parent/$oldName.child","plan/$parent/$newName.child") if(-f "plan/$parent/$oldName.child");   
    svnRename("plan/$parent","$oldName.child","$newName.child") if(-f "plan/$parent/$oldName.child");   
    #rename("plan/$parent/$oldName.disabled","plan/$parent/$newName.disabled") if(-f "plan/$parent/$oldName.disabled");   
    svnRename("plan/$parent","$oldName.disabled","$newName.disabled") if(-f "plan/$parent/$oldName.disabled");   
    }
  # Rename references in child plans
  my %children;
  getChildren($oldName,\%children);
  foreach my $child (keys %children){
    #rename("plan/$child/$oldName.parent","plan/$child/$newName.parent") if(-f "plan/$child/$oldName.parent");
    svnRename("plan/$child","$oldName.parent","$newName.parent") if(-f "plan/$child/$oldName.parent");
    #rename("plan/$child/$oldName.disabled","plan/$child/$newName.disabled") if(-f "plan/$child/$oldName.disabled");
    svnRename("plan/$child","$oldName.disabled","$newName.disabled") if(-f "plan/$child/$oldName.disabled");
    }
  # Rename files in the directory
  opendir(DIR,"plan/$oldName")||return("Unable to open 'plan/$oldName' directory");
  my @file=readdir(DIR);
  closedir(DIR);
  my $matchPlan="^$oldName\\.(.*)\$";
  my $matchCntl="^$oldName~\\.(.*)\$";
  foreach $file (@file){
    #rename("plan/$oldName/$oldName.$1","plan/$oldName/$newName.$1") if($file=~$matchPlan); 
    svnRename("plan/$oldName","$oldName.$1","$newName.$1") if($file=~$matchPlan); 
    #rename("plan/$oldName/$oldName~.$1","plan/$oldName/$newName~.$1") if($file=~$matchCntl); 
    svnRename("plan/$oldName","$oldName~.$1","$newName~.$1") if($file=~$matchCntl); 
    }
  # Rename plan directory
  #rename("plan/$oldName","plan/$newName");
  svnRename("plan","$oldName","$newName");
  return(rebuildPlan("$newName"));
  }

# Delete a plan
#
# Returns:  $msg - null if successful
#
sub deletePlan{
  my ($planName)=@_;
  return("Plan '$planName' not found.") if(! -d "plan/$planName");
  # Delete references in parent plans
  my %parents;
  getParents($planName,\%parents);
  foreach my $parent (keys %parents){
    unlink("plan/$parent/$planName.child") if(-f "plan/$parent/$planName.child");
    }
  # Delete references in child plans
  my %children;
  getChildren($planName,\%children);
  foreach my $child (keys %children){
    unlink("plan/$child/$planName.parent") if(-f "plan/$child/$planName.parent");
    }
  # Delete plan directory
  system("rm -fr plan/$planName");
  return('');
  }

# create or remove file used for plan state and relationships.

sub createFile{
  my ($file)=@_;
  return('') if(-f "$file");
  return("Unabled to create $file") if(!open(FILE,">$file"));
  close(FILE);
  return('');
  }

sub removeFile{
  my ($file)=@_;
  return('') if(! -f "$file");
  return("Unable to remove $file") if(!unlink("$file"));
  }

# Set the state of a plan 

sub setPlanState{
  my ($plan,$state)=@_;
  if($state eq 'enabled'){
    $msg=removeFile("plan/$plan/$plan.disabled");
    return($msg) if($msg);
    }
  elsif($state eq 'disabled'){
    $msg=createFile("plan/$plan/$plan.disabled");
    return($msg) if($msg);
    }
  else{return("State '$state' not recognized.");}
  return('');
  }

# Set the state of a relationship

sub setRelationState{
  my ($plan,$state,$relative,$relationship)=@_;
  if(!$relationship){
    $relationship='child' if(-f "plan/$plan/$relative.child");
    $relationship='parent' if(-f "plan/$plan/$relative.parent");
    return("Must specify a relationship") if(!$relationship);
    }
  if($relationship eq 'self'){
    if($state eq 'enabled'){$msg=removeFile("plan/$plan/$relative.disabled");}
    elsif($state eq 'disabled'){$msg=createFile("plan/$plan/$relative.disabled");}
    elsif($state eq 'removed'){$msg=deletePlan($plan);}
    return($msg);
    }
  my $reflection='parent';
  $reflection='child' if($relationship eq 'parent');
  if($state eq 'enabled' || $state eq 'disabled'){
    $msg=checkRelative($plan,$relative,$relationship);
    return($msg) if($msg);
    if($state eq 'enabled'){$msg=removeFile("plan/$plan/$relative.disabled");}
    else{$msg=createFile("plan/$plan/$relative.disabled");}
    return($msg) if($msg);
    $msg=createFile("plan/$plan/$relative.$relationship");
    return($msg) if($msg);
    $msg=createFile("plan/$relative/$plan.$reflection");
    return($msg) if($msg);
    }
  elsif($state eq 'removed'){
    $msg=removeFile("plan/$plan/$relative.$relationship");
    return($msg) if($msg);
    $msg=removeFile("plan/$plan/$relative.disabled");
    return($msg) if($msg);
    $msg=removeFile("plan/$relative/$plan.$reflection");
    return($msg) if($msg);
    $msg=removeFile("plan/$relative/$plan.disabled");
    return($msg) if($msg);
    }
  return('');
  }

sub getPlanner{
  my $userid='admin';     # This is a problem to be resolved
  my $email='root@localhost';  # should use configured admin address
  my $planner=new Planner($userid,$email); # Get a planner object to pass to scheme objects
  return($planner);
  }

# Get a table object for a plan using the table schema
#
# It would probably be best to modify Planner to get the userId and email
# on it's own.  The userid should come from a caboodle.conf, with an association
# to the Unix user.    e.g. AdminIds= account:id,account:id,...

sub getTable{
  my ($path)=@_;
  my $planner=getPlanner();
  return($planner->getTable($path));
  }

sub getConfig{
  my ($path)=@_;
  my $planner=getPlanner();
  my $planObj=new Plan($planner,$path);
  my $cfg=$planObj->cfg();
  return($cfg);
  }

# Compile plan
#   0 - if successful
#   '' - if not required
#   "..." if unsuccessful
#
sub compile{
  my ($plan)=@_;
  my $planner=getPlanner();
  my $planObj=new Plan($planner,$plan);
  my $compiler=$planObj->{COMPILER};
  return('') if(!$compiler);
  if(Config::install($plan,"adapter/$compiler.compiler")!=0){
    my $lcCompiler=lc($compiler);
    if(Config::install($plan,"adapter/$lcCompiler.compiler")!=0){
      return("CAB000E Unable to install adapter/$compiler.compiler");
      }
    else{
      $compiler=$lcCompiler;
      }
    }
  system("adapter/$compiler.compiler $plan");
  return("CAB000E Unsuccessful compile.") if($?);
  return(0);
  }

# Get compiler plan - called by plan compilers
#
# Synopsis: getCompilerPlan($plan,$compiler);
# Returns: 
#   ($msg) - error
#   ($msg,$heading,$planObj) - plan not using 'table' schema
#   ($msg,$heading,$planObj,$tableObj) - plan using 'table' schema
# 
sub getCompilerPlan{
  my ($plan,$compiler)=@_;
  return("CAB000E Plan parameter not found.") if(!$plan);
  my $planner=getPlanner();
  return("CAB000E Unable to instantiate planner.") if(!$planner);
  my $planObj=new Plan($planner,$plan);
  return("CAB000E Unable to instantiate plan.") if(!$plan);
  my $heading=$planObj->nb();
  return("CAB000E Unable to generate plan heading.") if(!$heading);
  return("CAB000E Plan does not use '$compiler' compiler.") if($compiler ne $planObj->{COMPILER});
  return('',$heading,$planObj) if($planObj->{SCHEMA} ne 'table');
  my $doc=$planObj->{DOC};
  my $tableDoc=$doc->child('plan')->child('table');
  my $tableObj=new Table($planner,$tableDoc);
  return('',$heading,$planObj,$tableObj); 
  } 

# Rebuild Plan
#   1) convert from *.xml to *.cfg
#   2) convert from *.cfg to *.xml
#   3) compile
#
sub rebuildPlan{
  my ($plan)=@_;
  my $cfg=Config::getConfig($plan);
  if(!open(CFG,">plan/$plan/$plan.cfg")){
    return(1);
    }
  print(CFG $cfg);
  close(CFG);
  system("bin/nbkit_cfg2xml $plan rebuild");  # rebuild option says don't mark as modified
  if($?!=0){
    return("Unable to rebuild plan $plan");
    }
  else{
    if(compile($plan)){
      return("Unable to compile plan $plan");
      }
    }
  if(!unlink("plan/$plan/$plan.cfg")){
    return("Unable to delete $plan.cfg");
    }
  return('');
  }

1
