#!/usr/bin/perl

# This class is used to manage the attributes element
#
#   <attributes>
#     <attribute id='' state='' time='' user='' reason=''
#
#        name     - Attribute name
#
#        value    - Value assigned to attribute
#                   For settings the value is always placed here.  For tables
#                   values are normally placed in row cells.  If a value is
#                   specified here, it overrides values in corresponding cells
#                   and the table column is not displayed.
#
#        type     - Type for code generation
#
#                   'string' - enclosed in quotes
#                   'text'   - not enclosed
#                   'number' - not enclosed
#                   'cell'   - not enclosed
#                   'code'   - not enclosed
#
#                   Looks like we can combine TYPE and FOR because anything
#                   other than VAR is going to be TEXT. All of the FOR values
#                   could just become types that are handled like TEXT.
#
#        for      - Specifies what the attribute is used for
#                   'var'   - variable for code generation (replacing 'arg')
#                             This is the default when not specified
#                   'arg'   - argument for assertion (deprecated)
#                   'doc'   - documentation
#                   'audit' - audit value to verify preceding 'var' or 'arg'
#                   'com'   - comment (how is this different than 'doc'?
#                   'opt'   - used for global options (not a table column)
#                   'code'  - raw code
#
#        format   - Input data modification
#                   ip - zero pad
#                   semi-colon separate set of values will be displayed a dropdown menu
#        audit    - identifies the audit program.  This enables an audit
#                   program to search through a plan and find values it
#                   is expected to audit.
#        default  - default value used for code generation when no value
#                   is specified.
#        term     - a NodeBrain term to use when generating code
#        tip      - deprecated - using title on update
#        title    - short descriptive title
#
#   </attributes>
#
# 2010-11-16 eat - included For value of 'code'
# 2011-03-26 eat - added onclick to image input for compatibility with FireFox 4.0 and IE

package Attributes;
use Carp;
use NBK::Planner::Planner;

#===========================================================================================================
# Scheme Methods

# Create new object

sub new{
  my ($package,$planner,$doc)=@_;
  bless {PLANNER => $planner, DOC => $doc} => $package;
  }

# We get to this update method via a section identifier
#
#    request=update;path=<planPath>:<type>.<element>[(<key>)]...;action=<action>;...
#
# This function is only called when the section identifier locates this COLUMNS element.
#
sub getUpdate{
  my ($self,$action,$upCol)=@_; 
  my $planner=$self->{PLANNER};
  my $msg=''; 
  if($action eq 'MoveUp' || $action eq 'MoveDown'){
    my @id=$planner->getParams('enable');
    foreach my $id (@id){
      $$upCol{$id}='selected';
      }
    @id=$planner->getParams('disable');
    foreach my $id (@id){
      $$upCol{$id}='selected';
      }
    }
  elsif($action eq 'Toggle' || $action eq 'Delete'){
    my $state='deleted';
    my @id=$planner->getParams('enable');
    $state='enabled' if($action eq 'Toggle');
    for my $id (@id){
      $$upCol{"$id"}=$state;
      }
    @id=$planner->getParams('disable');
    $state='disabled' if($action eq 'Toggle');
    for my $id (@id){
      $$upCol{"$id"}=$state;
      }
    @id=$planner->getParams('delete');
    for my $id (@id){
      $$upCol{"$id"}='deleted';
      }
    }
  elsif($action eq 'Update' || $action eq 'Add'){
    my $colId;
    if($action eq 'Update'){$colId=$planner->getParam('attribute');}
    else{$colId='-1';}
    my $colState=$planner->getParam('state');
    my $xml="<attribute id='$colId'";
    $xml.=" state='$colState'" if($colState ne 'enabled');
    my $utime=time();
    my $userId=$planner->getUserId();
    my $reason=$planner->getParam('reason');
    $xml.=" time='$utime' user='$userId' reason='$reason'";
    my $attrName=$planner->getParam('name');  # get name here because we need it later
    $xml.=" name='$attrName'";
    foreach my $attrName ('title','type','for','format','audit','value'){
      my $attrVal=$planner->getParam($attrName);
      $xml.=" $attrName='$attrVal'" if($attrVal ne '');
      }
    my $value=$planner->getParam('value');
    $value=$planner->encode($value);
    if($value){$xml.=">$value</attribute>\n";}
    else{$xml.="/>\n";}
    $$upCol{"$colId"}=$xml;
    }
  return($msg);
  }

# Convert a list document into XML text
#   Note: we are only interested here in the list part
#         the other plan elements are handled by Plan.pm
#
#   request=Update;path=<planPath>:<planName>:<type>.<element>[(<key>)]...;action=<action>;...
#   request=Update;path=<planPath>:<planName>:List.table;action=<action>;...
#
# Parameters obtained by calls to planner->getParam():
#
#   List.table.row*
#
#     action='Toggle' 
#
#       enabled=<rowid>;enabled=<rowid>;disabled=<rowid>;disabled=<rowid>;...
#
#     action='Update'
#
#       row=<rowId>;cells=<attributeNum>;cell1=<value>;cell2=<value>;cell3=<value>;...
#
# Update hash:
#   $upCol - $$upCol{$id}='enabled'|'disabled'|'deleted'|'<...new col XML...>'
#
# Returns string that is XML if it starts with '<', otherwise and error message

sub xml{
  my ($self,$path,$element,$action,$canEdit,$canCntl)=@_;
  my $planner=$self->{PLANNER};
  my $columns=$self->{DOC};  # prefer calling it $attributesDoc
  my %column;
  my %upCol;
  my $xml='';
  my $newId;
  my $cellMap='';
  my $assign=1 if($planner->getParam('action') eq 'Assign');
  
  # <attributes>
  $xml.="<attributes>\n";
  # support both new and old element name
  foreach my $elementName ('attribute','column'){
    foreach my $column ($columns->children($elementName)){
      next if(!$column);
      $column{$column->attribute('id')}=$column;
      }
    }
  # Get update hashes from parameters if there is an update for this element
  if($action && $element=~/^column/){
    return("User does not have control authority.") if(!$canCntl);
    $msg=$self->getUpdate($action,\%upCol) if($action);
    return($msg) if($msg);
    }
 # 2011-03-13 eat 0.8.5 - included column ordering
 if($action eq 'MoveUp' || $action eq 'MoveDown'){
    my @col;
    my %tab;
    my $target;
    my @mid;
    my @mcol;
    my $target;
    my @mid;
    if($action eq 'MoveUp'){($target,@mid)=sort {$a<=>$b} keys %upCol;}
    else{($target,@mid)=sort {$b<=>$a} keys %upCol;}
    if($target && @mid){
      foreach my $id (@mid){
        push(@mcol,$column{$id});
        delete $column{$id};
        }
      my $i=0;
      foreach my $id (sort {$a<=>$b} keys %column){
        $i++;
        if($upCol{$id}){
          if($id==$target){
            if($action eq 'MoveUp'){
              push(@col,@mcol);
              push(@col,$column{$id});
              }
            else{
              push(@col,$column{$id});
              push(@col,@mcol);
              }
            }
          # else ignore now because it is a moving column
          }
        else{push(@col,$column{$id});}
        }
      my $i=0;
      foreach my $col (@col){
        $i++;
        $tab{$i}=$col;
        $cellMap.=$col->attribute('id').'='.$i.';';
        }
      %column=%tab;
      }
    }
  foreach my $id (sort {$a<=>$b} keys %column){
    my $update=$upCol{$id};
    my $columnDoc=$column{$id};
    my $value=$columnDoc->value;
    $value=$columnDoc->attribute('value') if(!$value);
    if(!$update && $assign){   # other updates override assignment
      $update='assign';
      $newValue=$planner->getParam("val$id");
      }
    my $columnName=$columnDoc->attribute('name');
    if($update=~/^</){
      $xml.=$update;
      delete $upCol{$id};
      #$upCol{$id}='';
      $columnName='';
      $columnName=$1 if($update=~/name='(.*?)'/);
      $planner->postChange('PlanTableAttributeUpdate',$path,"($columnName,...)");
      }
    elsif($update eq 'deleted'){   # note: when a column is deleted corresponding cells will be deleted from table rows
      $planner->postChange('PlanTableAttributeDelete',$path,"($columnName,...)");
      }
    elsif($update eq 'assign' && $newValue ne $value){
      $state=$columnDoc->attribute('state');
      if($state eq 'disabled'){$xml.="<attribute id='$id' state='disabled'";}
      else{$xml.="<attribute id='$id'";}
      my $elementTime=time();
      my $elementUserId=$planner->getUserId();
      my $elementReason=$planner->getParam('reason');
      $xml.=" time='$elementTime' user='$elementUserId' reason='$elementReason'";
      $xml.=" name='$columnName'";
      foreach my $attrName ('title','type','format','for','audit','tip'){
        my $attrVal=$columnDoc->attribute($attrName);
        $xml.=" $attrName='$attrVal'" if($attrVal ne '');
        }
      #my $value=$columnDoc->attribute('value');
      #print("old value=$value<br>\n");
      #print("new value=$newValue<br>\n");
      $value=$planner->format($newValue,$columnDoc->attribute('format'));
      #print("formatted value=$value<br>\n");
      $value=$planner->encode($value);
      #print("encoded value=$value<br>\n");
      if($value){$xml.=">$value</attribute>\n";}
      else{$xml.="/>\n";}
      $planner->postChange('PlanTableAttributeAssign',$path,"($columnName=$value)");
      }
    elsif($update eq 'enabled' || $update eq 'disabled'){
      if($update eq 'disabled'){
        $xml.="<attribute id='$id' state='disabled'";
        $verb='Disable';
        }
      else{
        $xml.="<attribute id='$id'";
        $verb='Enable';
        }
      my $elementTime=time();
      my $elementUserId=$planner->getUserId();
      my $elementReason=$planner->getParam('reason');
      $xml.=" time='$elementTime' user='$elementUserId' reason='$elementReason'"; 
      $xml.=" name='$columnName'";
      foreach my $attrName ('title','type','format','for','audit','tip'){
        my $attrVal=$columnDoc->attribute($attrName);
        $xml.=" $attrName='$attrVal'" if($attrVal ne '');
        }
      my $value=$columnDoc->value;
      $value=$columnDoc->attribute('value') if(!$value); # old style
      if($value){$xml.=">$value</attribute>\n";}
      else{$xml.="/>\n";}
      $planner->postChange('PlanTableAttribute'.$verb,$path,"($columnName,...)");
      }
    else{
      my $terminator="/>\n";
      $terminator=">$value</attribute>\n" if($value);
      my $element=$planner->getElementXml('attribute',$column{$id},'id,state,time,user,reason,name,title.tip,type,format,audit,for',$terminator);
      $element=$1."id='$id'".$2 if($element=~/^(.*?)id='\d+'(.*)$/);
      $xml.=$element."\n";
      #$xml.=$planner->getElementXml('attribute',$column{$id},'id,state,time,user,reason,name,title.tip,type,format,audit,for',$terminator);
      }
    $newId=$id;
    }
  # handle additions
  foreach my $id (sort {$a<=>$b} keys %upCol){
    if($id<0 && $upCol{$id}=~/^</){
      my $colStr=$upCol{$id};
      $newId++;
      $colStr=~s/^<attribute id='-\d+'/<attribute id='$newId'/g;
      $xml.=$colStr;
      $attrName='';
      $attrName=$1 if($colStr=~/name='(.*?)'/);
      $planner->postChange('PlanTableAttributeAdd',$path,"($attrName,...)");
      }
    }
  $xml.="</attributes>\n";
  return($xml,$cellMap);  
  }


# Return HTML representation
#
#   Path  - portion of section consumed by parent element
#           This is required for constructing URL's to get
#           back for update. 
#
#   Action  Element
#   ------  --------
#
#   'View'  columns
#           column(<colId>)
#
#   'Edit'  columns
#           column(<colId>)
#   
sub html{
  my ($self,$path,$element,$action,$canEdit,$canCntl)=@_;
  my $planner=$self->{PLANNER};
  my $columns=$self->{DOC};
  $columnId=$1 if($element=~/^column\((\d+)\)$/);
  $action='Display' if(!$action);  # default action
  $element='columns' if(!$element);
  my $html='';
  my %column;
  my $colnum=0;
  # support both new and old element name
  foreach my $elementName ('attribute','column'){
    foreach my $column ($columns->children($elementName)){
      next if(!$column);
      $column{$column->attribute('id')}=$column;
      $colnum++;
      }
    }
  # List.columns
  if($action eq 'Edit' && $canCntl){  # If edit requested and authorized
    $html.="<form style='margin-bottom:0;' name=Attributes action='Planner.cgi' method=post>\n";
    $html.="<input type='hidden' name='request' value='Update'/>\n";
    $html.="<input type='hidden' name='action' value=''/>\n";
    $html.="<input type=hidden name='path' value='$path.columns'/>\n";
    $html.="<input type=hidden name='attribute' value='$columnId'/>\n" if($columnId);
    my ($columnDoc,$columnName,$columnTitle,$columnType,$columnFormat,$columnAudit,$columnValue,$columnFor);
    if($columnId){
      $columnDoc=$column{$columnId};
      if($columnDoc){
        $columnName=$columnDoc->attribute('name');
        $columnTitle=$columnDoc->attribute('title');
        $columnTitle=$columnDoc->attribute('tip') if(!defined $columnTitle); # old keyword
        $columnType=$columnDoc->attribute('type');
        $columnFor=$columnDoc->attribute('for');
        $columnFormat=$columnDoc->attribute('format');
        $columnAudit=$columnDoc->attribute('audit');
        #$columnDefault=$columnDoc->attribute('default');
        #$columnValue=$columnDoc->attribute('value');
        $columnValue=$columnDoc->value;
        }
      }
    $html.="<table width='100%' class='fill'><tr><th>Attribute</th><th>";
    $html.="<select name='state'>\n";
    foreach my $state ('enabled','disabled'){
      my $selected='';
      $selected=' SELECTED' if($state eq $rowState);
      $html.="<option value='$state'$selected>$state</option>\n";
      }
    $html.="</select>\n";
    $html.="&nbsp; <input type=image src='planAdd.gif' title='Create new column' name='image' onclick=\"this.form.action.value='Add'\"/>\n";
    $html.="&nbsp; <input type=image src='planUpdate.gif' title='Write changes to selected column' name='image' onclick=\"this.form.action.value='Update'\"/>\n" if($columnDoc);
    $html.="</td></tr>\n";
    $html.="<tr><th width='5%'>Parameter</th><th>Value</th><tr>\n";
    # id,name,value,type,for,format,audit,size,default,term,state,comment,tip,title
    $html.="<tr><td><b>Name</b></td><td><input type=text name=name value='$columnName' size=80> Short attribute name without spaces or special characters.</td><tr>";
    $html.="<tr><td><b>Title</b></td><td><input type=text name=title value='$columnTitle' size=80> Short title for tips like this.</td><tr>";
    $html.="<tr><td><b>Type</b></td><td><select name=type>\n";
    foreach my $option ('string','text','number','cell','code'){
      my $selected='';
      $selected=' SELECTED' if($option eq $columnType);
      $html.="<option value='$option'$selected>$option</option>\n";
      }
    $html.="</select> Type for NodeBrain code generation.  Use [string] if unsure when adding new attributes.</td><tr>";
    $html.="<tr><td><b>For</b></td><td><select name=for>\n";
    foreach my $option ('arg','doc','audit','value','opt','code'){
      my $selected='';
      $selected=' SELECTED' if($option eq $columnFor);
      $html.="<option value='$option'$selected>$option</option>\n";
      }
    $html.="</select> This directs code generation.  User [arg] for attributes to match and [doc] for others.</td></tr>";
    $html.="<tr><td><b>Format</b></td><td><input type=text name=format value='$columnFormat' size=80> Format of input. Use 'ip' for IP addresses.</td></tr>";
    $html.="<tr><td><b>Audit</b></td><td><input type=text name=audit value='$columnAudit' size=80> Audit tag for audit programs.  Leave blank if unsure.</td></tr>";
    $html.="<tr><td><b>Value</b></td><td>";
    if($columnType eq 'code'){
      $html.="<textarea name=value rows=10 cols=120>$columnValue</textarea>";
      }
    else{
      $html.="<input type=text name=value value='$columnValue' size=80>";
      }
    $html.=" Constant value used in all rows.  Leave blank if unsure.</td></tr>";
    $html.="<tr><th>Reason</th><th colspan=2><input type=text name=reason size=80> Change request number and/or short reason for change.</th></tr>\n";
    $html.="</table>\n";
    }
  $html.="<table width='100%' border=0 cellpadding=3 cellspacing=1><tr><th width='20'>\n";
  if($action eq 'Edit' && $canCntl){
    $html.="<input type=image src='planUp.gif' name='image' onclick=\"this.form.action.value='MoveUp'\" title='Move selected rows above the first selected row.'/>";
    $html.="<input type=image src='planDown.gif' name='image' onclick=\"this.form.action.value='MoveDown'\" title='Move selected rows below the last selected row.'/>\n";

    $html.="</th><th width='37'>";
    $html.="<input type=image src='planToggle.gif' name='image' onclick=\"this.form.action.value='Toggle'\" title='Toggle state of selected attribute'>\n";
    $html.="<input type=image src='planCut.gif' name='image' onclick=\"this.form.action.value='Delete'\" title='Delete selected attribute'/></th>\n";
    }
  else{$html.="</th><th width='19'></th>";}
  $html.="<th>Attribute</th><th>Title / Value</th><th>Type</th><th>For</th><th>Format</th><th>Audit</th><th>Revised</th><th>Planner</th><th>Reason</th></tr>\n";
  $row=0;
  foreach my $id (sort {$a<=>$b} keys %column){
    my $colDoc=$column{$id};
    my $state=$column{$id}->attribute('state');
    $colName=$column{$id}->attribute('name');
    $colType=$column{$id}->attribute('type');
    $colFormat=$column{$id}->attribute('format');
    $colFormat=join("<br>",split(/;/,$colFormat)) if($colFormat=~/;/);
    $colFor=$colDoc->attribute('for');
    $colAudit=$colDoc->attribute('audit');
    #$colDefault=$column{$id}->attribute('default');
    $colTitle=$column{$id}->attribute('title');
    $colTitle=$column{$id}->attribute('tip') if(!$colTitle); # support old format
    $colValue=$column{$id}->value;  # get value
    $colValue=$column{$id}->attribute('value') if(!$colValue); # get old style if no value
    $colValue=$column{$id}->attribute('default') if(!$colValue); # get old style default if no value
    my $elementTime=$colDoc->attribute('time');
    my $elementDate=$planner->getUserDate($elementTime) if($elementTime);
    my $elementUserId=$colDoc->attribute('user');
    my $elementUserName='';
    if($elementUserId){
      my $userObj=$planner->getUser($elementUserId);
      $elementUserName=$userObj->getName();
      }
    #my $elementUserName=$planner->getUserName($elementUserId) if($elementUserId);
    my $elementReason=$colDoc->attribute('reason');
    $class='even';
    $class='odd' if($row%2);
    $class='' if($id eq $columnId);
    $html.="<tr class=$class><td>$id</td>";
    my $linkRequest='Display';
    my $linkTitle='Select to highlight this entry';
    if($canCntl){
      $linkRequest='Edit';
      $linkTitle='Select to edit this entry';
      }
    if($state eq 'deleted'){
      $fontOn="<font color='#b0b0b0'>";  # should use style instead
      $fontOff="</font>";
      $stateImage='deleted.gif';
      $checkName='enable';
      }
    elsif($state eq 'disabled'){
      $fontOn="<font color='#b0b0b0'>";
      $fontOff="</font>";
      $stateImage='planDisabled.gif';
      $checkName='enable';
      }
    else{
      $fontOn="";
      $fontOff="";
      $stateImage='planEnabled.gif';
      $checkName='disable';
      }

    $html.="<td><a href='Planner.cgi?request=$linkRequest;path=$path.column($id)'><img src='$stateImage' border=0 title='$linkTitle'></a>";
    $html.="<input type=checkbox name='$checkName' value='$id'>" if($action eq 'Edit' && $canCntl);
    $html.="</td>";
    $html.="<td>$fontOn$colName$fontOff</td>";
    $html.="<td>$fontOn$colTitle$fontOff<br>$fontOn<b>$colValue</b>$fontOff</td>";
    foreach my $value ($colType,$colFor,$colFormat,$colAudit){
       $html.="<td>$fontOn$value$fontOff</td>";
       }
    $html.="<td class=fill>$elementDate</td><td class=fill>$elementUserName</td><td class=fill>$elementReason</td></tr>\n";
    $row++;
    }
  $html.="</table>";
  $html.="</form>" if($action eq 'Edit' && $canCntl);
  return($html);
  }

sub nb{
  my ($self)=@_;
  my $columnsDoc=$self->{DOC};
  my $nb="# Attributes:\n#\n";
  return($nb) if(!$columnsDoc);
  my %column;
  my %columnType;
  # support both new and old element name
  foreach my $elementName ('attribute','column'){
    foreach my $columnDoc ($columnsDoc->children($elementName)){
      next if(!$columnDoc);
      $columnId=$columnDoc->attribute('id');
      $columnElement{$columnId}=$columnDoc;      # create hash to order and lookup by id
      }
    }
  my $dataCol=0;
  my $dataColId;
  my $commentCol=0;
  foreach my $columnId (sort {$a<=>$b} keys %columnElement){
    $element=$columnElement{$columnId};
    my $columnState=$element->attribute('state');
    next if($columnState && $columnState ne 'enabled');
    $columnName=$element->attribute('name');
    $columnFor=$element->attribute('for');
    $columnFor='arg' if(!$columnFor);
    $columnType=$element->attribute('type');
    $columnType='cell' if(!$columnType);       # default type is cell expresion
    $columnTerm=$element->attribute('term');
    $columnComment=$element->attribute('comment');
    $columnComment="# $columnComment" if($columnComment);
    if($columnFor=~/^(comment|audit|doc)$/){
      $commentCol++;
      $dataColId=" #)";
      }
    else{
      $dataCol++;
      $dataColId=sprintf("%2.1s)",$dataCol);
      }
    $nb.="#    $dataColId $columnName ($columnType,$columnFor) $columnComment\n";
    }
  $nb.="#\n";
  return($nb);
  }

sub cfg{
  my ($self)=@_;
  my $columnsDoc=$self->{DOC};
  return($nb) if(!$columnsDoc);
  my $nb='';
  my %column;
  my %columnType;
  # support both new and old element name
  foreach my $elementName ('attribute','column'){
    foreach my $columnDoc ($columnsDoc->children($elementName)){
      next if(!$columnDoc);
      $columnId=$columnDoc->attribute('id');
      $columnElement{$columnId}=$columnDoc;      # create hash to order and lookup by id
      }
    }
  my %default;
  #$default{type}='string';
  #$default{for}='arg';
  foreach my $columnId (sort {$a<=>$b} keys %columnElement){
    $element=$columnElement{$columnId};
    my $columnState=$element->attribute('state');
    my $stateFlag='.';
    $stateFlag='!' if($columnState eq 'disabled');
    $columnName=$element->attribute('name');
    $nb.="\n$stateFlag$columnName";
    my $value=$element->value;
    $value=$1 if($value=~/^\r?\n(.*)$/s);
    $value=$element->attribute('value') if(!$value); # old style
    $value=$element->attribute('default') if(!$value); # old style
    if($value){
      my @text=split(/\r?\n/,$value);
      if($#text>0){
        $nb.="{\n";
        foreach my $text (@text){
          $nb.=" $text\n";
          }
        $nb.='}';
        }
      else{$nb.=": $value";}
      }
    $nb.="\n";
    # handle all meta attributes except "name" and "value"
    foreach my $metaAttr ('title','type','for','format','audit','comment'){
      my $metaValue=$element->attribute($metaAttr);
      #next if($metaAttr eq 'for' && $value);
      $nb.=" $metaAttr: $metaValue\n" if($metaValue && $metaValue ne $default{$metaAttr});
      }
    }
  $nb.="\n>\n";
  foreach my $columnId (sort {$a<=>$b} keys %columnElement){
    $element=$columnElement{$columnId};
    next if($element->attribute('for') eq 'opt');
    my $value=$element->value;
    $value=$element->attribute('value') if(!$value);  # old style
    $value=$element->attribute('default') if(!$value);  # old style
    $name=$element->attribute('name');
    $nb.=" $name: $value\n";
    } 
  return($nb);
  }

1;
