#!/usr/bin/perl

# This class provides support to all planner classes
#==============================================================================
# Change History
# 2011-03-04 eat 0.8.0 - included link method
# 2011-10-12 eat 0.8.1 - made sure we have "guest" and default to guest for unknown userid's
# 2011-10-23 eat 0.8.1 - encode/decode on copyright symbol
# 2012-04-07 eat 0.8.1 - changed encoding of ' from &apos; to &#39;
#==============================================================================

package Planner;
use Carp;
use XML::Parser;
use XML::SimpleObject;
use CGI qw(:standard);

use NBK::Planner::User;
use NBK::Planner::Plan;
use NBK::Planner::Table;

# Construct new planner

sub new{
  my ($package,$UserId,$UserEmail)=@_;
  my %hash;
  my $parser=new XML::Parser(Style => 'Tree');
  $hash{PARSER}=$parser;
  $hash{USERID}=$UserId;
  $hash{USERROLE}='guest';
  $hash{USERNAME}='guest';
  $hash{USEREMAIL}=$UserEmail;
  my $doc=new XML::SimpleObject($parser->parsefile("plan/CaboodleAccess/CaboodleAccess.xml"));
  my $planDoc=$doc->child('plan');
  my $tableDoc=$planDoc->child('table');
  my $rowsDoc=$tableDoc->child('rows');
  $hash{USER}->{guest}=new User("Guest","guest","root\@localhost");
  foreach my $rowDoc ($rowsDoc->children('row')){
    my $userId,$userRole,$userName,$userEmail;
    foreach my $cellDoc ($rowDoc->children('cell')){
      $userId=$cellDoc->value if($cellDoc->attribute('id') eq '1');
      $userRole=$cellDoc->value if($cellDoc->attribute('id') eq '2');
      $userName=$cellDoc->value if($cellDoc->attribute('id') eq '3');
      $userEmail=$cellDoc->value if($cellDoc->attribute('id') eq '4');
      }
    $hash{USER}->{$userId}=new User($userName,$userRole,$userEmail);
    if($userId eq $UserId){
      $hash{USERROLE}=$userRole;
      $hash{USERNAME}=$userName;
      $hash{USEREMAIL}=$userEmail;
      }
    }
  $hash{CHANGE}=[]; # list of changes reported by element classes
  $hash{EVENT}=[];  # list of events reported by element classes
  bless \%hash => $package;
  }

# Method to post document changes
#
# Verb: Add, Update, Delete
# Type: Row, Column, Permit, Action
# Path: Path to element
#       e.g. Home/Planner/MyTab:List.table(fred)
# Element: Specific element identifier
#       e.g. row(5)
# Text: text for change history 

sub postChange{
  my ($self,$type,$path,$text)=@_;
  my $plan=$path;
  $plan=$1 if($plan=~/^(.*?)\.(.*)$/);
  $plan=$2 if($plan=~/(.*?\/)*(.*?)$/);
  push(@{$self->{EVENT}},[$type,$plan,$text]);
  }

sub alertChange{
  my ($self,$file)=@_;
  open(LOG,">>log/Planner.log");
  open(NB,">>log/Planner.nb");
  foreach my $list (@{$self->{EVENT}}){
    my ($type,$plan,$text)=@$list;
    my ($sec,$min,$hour,$mday,$mon,$year)=gmtime();
    my $date=sprintf("%4.4d-%2.2d-%2.2d %2.2d:%2.2d:%2.2d",$year+1900,$mon+1,$mday,$hour,$min,$sec);
    print(LOG "$date $type $plan $text\n");
    print(NB "alert type=\"$type\",plan=\"$plan\",text=\"$text\";\n");
    }
  close(NB);
  close(NB);
  }

# format a value based on format specification

sub format{
  my ($self,$value,$format)=@_;
  return($value) if($format=~/;/);  # menu format - use value as is
  if($format eq 'ip'){
    $value=sprintf("%3.3d.%3.3d.%3.3d.%3.3d",$1,$2,$3,$4) if($value=~/^(\d+)\.(\d+).(\d+).(\d+)$/);
    }
  return($value);
  }

# Methods to encode XML text

sub encodeAmp{     
  my ($self,$text)=@_;
  $text=~s/&/&amp;/g;
  return($text);
  }

sub encode{
  my ($self,$text)=@_;
  $text=~s/&/&amp;/g;
  $text=~s/</&lt;/g;
  $text=~s/>/&gt;/g;
  #$text=~s/'/&apos;/g;
  $text=~s/'/&#39;/g;
  $text=~s/"/&quot;/g;
  $text=~s/\xA9/&#169;/g;
  return($text);
  }

sub decode{
  my ($self,$text)=@_;
  $text=~s/&amp;/&/g;
  $text=~s/&lt;/</g;
  $text=~s/&gt;/>/g;
  $text=~s/&apos;/'/g;
  $text=~s/&quot;/"/g;
  $text=~s/&#169;/\xA9/g;
  return($text);
  }

# Method to create link to another plan
sub link{
  my ($self,$path,$text)=@_;
  my $linktext='';
  while($text=~/^(.*?) plan\/(\w+?)\/(.*)$/s){
    #print("in:$text<br>\n");
    $linktext.=$1." <a href='Planner.cgi?request=Display;path=$path/$2'>plan/$2</a>/";
    $text=$3;
    }
  $linktext.=$text;
  return($linktext);
  }

# Method to get user object

sub getUser{
  my ($self,$userid)=@_;
  if($self->{USER}->{$userid}) {return($self->{USER}->{$userid});}
  return($self->{USER}->{guest});
  }

# Methods to get current user information

sub getUserId{
  my ($self)=@_;
  return($self->{USERID});
  }

sub getUserName{
  my ($self)=@_;
  return($self->{USERNAME});
  }

sub getUserEmail{
  my ($self)=@_;
  return($self->{USEREMAIL});
  }

sub getUserRole{
  my ($self)=@_;
  return($self->{USERROLE});
  }

sub getPlanners{
  my ($self)=@_;
  my @planner;
  foreach my $userId (keys %{$self->{USER}}){
    $userObj=$self->{USER}->{$userId};
    push(@planner,$userId) if($userObj->getRole()=~/^(planner|admin)$/);
    }
  return(@planner);
  }

sub getPlannerMenu{
  my ($self)=@_;
  my %planner;
  foreach my $userId (keys %{$self->{USER}}){
    $userObj=$self->{USER}->{$userId};
    $planner{$userObj->getName()}=$userId if($userObj->getRole()=~/^(planner|admin)$/);
    }
  my $html="<select name=user>\n";
  foreach my $userName (sort keys %planner){
    $html.="<option value='$planner{$userName}'>$userName</option>\n"; 
    }
  $html.="</select>\n";
  return($html);
  }

    

# User time requires access to user profile

sub getUserTime{
  my ($self,$utime)=@_;
  # Need to include logic for user time zone offset
  my ($sec,$min,$hour,$mday,$mon,$year)=gmtime($utime);
  my $userTime=sprintf("%4.4d-%2.2d-%2.2d %2.2d:%2.2d:%2.2d",$year+1900,$mon+1,$mday,$hour,$min,$sec);
  return($userTime);
  }

sub getUserDate{
  my ($self,$utime)=@_;
  # Need to include logic for user time zone offset
  my ($sec,$min,$hour,$mday,$mon,$year)=gmtime($utime);
  my $userDate=sprintf("%4.4d-%2.2d-%2.2d",$year+1900,$mon+1,$mday);
  return($userDate);
  }

# Load a plan document using a plan path and name 

# We are flexible here on how the plan file is specified until all the
# code settles down.
#
#         name
#         name~
#   path  name
#   path  name~
#   path  ~
#   path
#   path~

sub loadPlan{
  my ($self,$planPath)=@_;
  my $planPart=$planPath;
  $planPart=$2 if($planPath=~/(.*?\/)+([^\/]+)$/);
  my $plan=$planPart;
  $plan=$1 if($plan=~/(.*?)~/);
  my $file="plan/$plan/$planPart.xml";
  return(0) if(! -f "$file");
  my $doc=new XML::SimpleObject($self->{PARSER}->parsefile($file));
  return($doc);
  }

sub loadDoc{
  my ($self,$file)=@_;
  return(0) if(! -f "$file");
  my $doc=new XML::SimpleObject($self->{PARSER}->parsefile($file));
  return($doc);
  }

# Get CGI parameter
#   This provides a place to make things work in batch mode also.
#   Just need to provide a way for batch utilities to supply the parameters
#   and need a $self->{MODE} to know we need to use it
#
sub getParam{
  my ($self,$name)=@_;
  my $value=param($name);
  return($value);
  }

sub getParams{
  my ($self,$name)=@_;
  my @value=param($name);
  return(@value);
  }

# Create a string representation of an element open tag 
#
#   $element     - name of element
#   $doc         - structure containing the element
#   $attributes  - comma separated list of attribute used to control attribute order in the output
#                  May want to include a special symbol prefix to indicate deletion
#   $close       - '/>', '>', or '' (use '' if you want to add new attributes)
#
sub getElementXml{
  my ($self,$element,$doc,$attributes,$close)=@_;
  my $xml="<$element";
  @attribute=split(/,/,$attributes);
  foreach $attribute (@attribute){
    my ($prefName,@deprecatedName)=split(/\./,$attribute);
    my $value=$doc->attribute($prefName);
    if(!defined $value){
      foreach my $deprecatedName (@deprecatedName){
        $value=$doc->attribute($deprecatedName);
        last if(defined $value);
        }
      }
    $xml.=" $prefName='$value'" if($value ne '');
    }
  $xml.=$close;
  return($xml);
  }

# Get Plan State - we manage this external from the xml now

sub getPlanState{
  my ($self,$path)=@_;
  my ($path,$element)=split(/:/,$path);
  ($path,$element)=($1,$2) if($path=~/^(.*?)\.(.*)$/);
  my $plan=$filePath;
  $plan=$2 if($plan=~/(.*\/)+([^\/]+)$/);
  my $state='enabled';
  $state='disabled' if(-f "plan/$plan/$plan.disabled");
  return($state);
  }

# This is a replacement for getPlanParts
#   Here we use a full file name to allow access to control plans
#   in kit directories as well as caboodle directories
#   A hash is used to return information so it is easy to add
#   new items later.
#
#    title
#    description
#    type
#    admins
#
# Returns '' on success, otherwise error message
#
sub getCntl{
  my ($self,$planName,$hash,$dir)=@_;
  my $file="plan/$planName/$planName~.xml";
  $file="$dir/$file" if($dir && $dir ne '.');
  my $doc=$self->loadDoc($file);
  return("Unable to load file") if(!$doc);
  my $plan=$doc->child('plan');
  $$hash{scheme}='Unknown';
  $$hash{schema}='Unknown';
  $$hash{type}='Unknown';
  my $planDescObj=$plan->child('description');
  $$hash{title}=$planDescObj->attribute('title');
  $$hash{purpose}=$planDescObj->value;
  my $planAdmins='';
  my $controlObj=$plan->child('control');
  if($controlObj){
    $$hash{scheme}=$controlObj->attribute('scheme');
    $$hash{scheme}=$controlObj->attribute('type') if(!$$hash{scheme});
    $$hash{type}=$$hash{scheme}; # enable transition to scheme from type
    $$hash{schema}=$controlObj->attribute('schema');
    my $planPermitsObj=$controlObj->child('permits');
    if($planPermitsObj){
      foreach $element ($planPermitsObj->children('permit')){
        $userId=$element->attribute('user');
        $role=$element->attribute('role');
        $admin=$self->getUser($userId);
        if($admin){
          $planAdmins.=$admin->getName();
          $planAdmins.=" ($role)" if($role);
          $planAdmins.=', ';
          }
        }
      }
    $planAdmins=$1 if($planAdmins=~/^(.*), $/);
    $$hash{admins}=$planAdmins;
    }
  return('');
  }

# this can move to Plan.pm perhaps
# This should normally be called for control plans 
sub getPlanParts{
  my ($self,$planPath)=@_;
  my $doc=$self->loadPlan($planPath.'~');
  if(!$doc){
    my ($filePath,$element)=split(/:/,$planPath);
    ($filePath,$element)=($1,$2) if($filePath=~/^(.*?)\.(.*)$/);
    my $planName=$filePath;
    $planName=$2 if($planName=~/(.*\/)+([^\/]+)$/);
    return($planName,'Missing Plan','This plan does not exist','Unknown','missing','None');
    }
  my $plan=$doc->child('plan');
  my $planType='Unknown';
  my $planState=getPlanState($planPath);
  my $planDescObj=$plan->child('description');
  my $planName=$planDescObj->attribute('name');
  my $planTitle=$planDescObj->attribute('title');
  my $planDesc=$planDescObj->value;
  my $planAdmins='';
  my $controlObj=$plan->child('control');
  if($controlObj){
    $planType=$controlObj->attribute('scheme');
    $planType=$controlObj->attribute('type') if(!$planType);
    my $planPermitsObj=$controlObj->child('permits');
    if($planPermitsObj){
      foreach $element ($planPermitsObj->children('permit')){
        $userId=$element->attribute('user');
        $role=$element->attribute('role');
        $admin=$self->getUser($userId);
        if($admin){
          $planAdmins.=$admin->getName();
          $planAdmins.=" ($role)" if($role);
          $planAdmins.=', ';
          }
        }
      }
    $planAdmins=$1 if($planAdmins=~/^(.*), $/);
    }
  return($planName,$planTitle,$planDesc,$planType,$planState,$planAdmins);
  }

# Get table object for a plan using the table schema
#
#   This is provided as a short cut for plan compilers based on simple tables
#   that are used similar to the original Sysmon configuration files.  The Table
#   object provides an "extract" function that returns rows in a structure that
#   is simple for compilers to work with.
#
sub getTable{
  my ($self,$path)=@_;
  my $planObj=new Plan($self,$path);
  my $doc=$planObj->{DOC};
  my $tableDoc=$doc->child('plan')->child('table');
  return(0) if(!$tableDoc);
  my $tableObj=new Table($self,$tableDoc);
  return($tableObj);
  }

1; 
